import os
import pytest
from pathlib import Path

from beaker_bunsen.corpus.resources import Resource
from beaker_bunsen.corpus.vector_stores.chromadb_store import ChromaDBLocalStore
from beaker_bunsen.corpus.embedders.embedder import Embedder
from beaker_bunsen.corpus.loaders.local_file_loader import LocalFileLoader

# @pytest.fixture()
# def chromadb_store_path(tmp_path_factory):
#     return str(tmp_path_factory.mktemp("test"))


# @pytest.fixture()
# def test_data_paths():
#     data = Path(__file__).parent / "data"
#     return [data / "documents", data / "examples"]


# def test_base_embedder(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=test_data_paths)
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#     )

#     records_before_ingestion = store.get_all()
#     embedder.ingest()
#     ingested_records = store.get_all()

#     assert len(records_before_ingestion) == 0
#     assert len(ingested_records) > 0


# def test_embedder_ingest_with_metadata(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=test_data_paths)
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#     )

#     embedder.ingest(metadata={"embedder": "test_metadata"})

#     ingested_records = store.get_all()
#     metadata_map = {os.path.basename(record.uri): record.metadata for record in ingested_records}

#     assert all('embedder' in record.metadata.keys() for record in ingested_records)
#     assert metadata_map["Lunar_Sample_Laboratory_Facility.html"] == {"embedder": "test_metadata"}
#     assert metadata_map["yorkshire.txt"] == {"embedder": "test_metadata", "source": "wikipedia", "url": "https://en.wikipedia.org/wiki/Yorkshire"}


# def test_embedder_ingest_location_override(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=[test_data_paths[0]])
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#     )

#     embedder.ingest(locations=[test_data_paths[1]])
#     ingested_records = store.get_all()

#     default_loader_uris = [record.uri for record in loader.discover()]
#     ingested_uris = [record.uri for record in ingested_records]

#     assert len(ingested_records) > 0
#     assert ingested_uris != default_loader_uris


# def test_embedder_ingest_partition_override(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=test_data_paths)
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#     )

#     embedder.ingest(partition="other")

#     records_in_default_partition = store.get_all()
#     records_in_other_partition = store.get_all(partition="other")

#     assert len(records_in_default_partition) == 0
#     assert len(records_in_other_partition) > 0


# def generate_embedding_function(start=0):
#     glob = {"count": start}
#     def embedding_function(resource: Resource, outer_ref=glob):
#         return_val = [float(outer_ref["count"])] * 384
#         outer_ref["count"] += 1
#         return return_val
#     return embedding_function


# def test_embedder_embedding_function(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=test_data_paths)
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#         embedding_function=generate_embedding_function(1)
#     )

#     embedder.ingest()

#     ingested_records = store.get_all(include_embeddings=True)
#     first_five_embedding_values = sorted([record.embedding[0] for record in ingested_records])[:5]

#     assert first_five_embedding_values == [1.0, 2.0, 3.0, 4.0, 5.0]


# def test_embedder_embedding_function_override(chromadb_store_path, test_data_paths):
#     store = ChromaDBLocalStore(path=chromadb_store_path)
#     loader = LocalFileLoader(locations=test_data_paths)
#     embedder = Embedder(
#         loader=loader,
#         store=store,
#     )

#     embedder.ingest(
#         embedding_function=generate_embedding_function(10)
#     )

#     ingested_records = store.get_all(include_embeddings=True)
#     first_five_embedding_values = sorted([record.embedding[0] for record in ingested_records])[:5]

#     assert first_five_embedding_values == [10.0, 11.0, 12.0, 13.0, 14.0]



from pathlib import Path
import hashlib


def find_pyproject_file() -> Path | None:
    cwd = Path.cwd()
    for path in [cwd, *cwd.parents]:
        potential_file = path / "pyproject.toml"
        if potential_file.is_file():
            return potential_file
    return None


def calculate_content_hash(
    content: str,
    hash_algo: str = "sha256"
) -> str:
    if isinstance(content, str):
        content = content.encode()
    hash_obj = hashlib.new(hash_algo, data=content)
    return str(hash_obj.hexdigest())



import importlib
import inspect
import os
import os.path
import sys
import zipfile
from abc import ABC, abstractmethod
from functools import cache
from pathlib import Path
from typing import Any, TYPE_CHECKING

from ..types import URI
if TYPE_CHECKING:
    from .. import Corpus
    from .base import BaseLoader


def _is_scheme(obj: Any):
    return bool(
        inspect.isclass(obj)
        and issubclass(obj, Scheme)
        and callable(getattr(obj, "read", None))
    )


@cache
def all_schemes():
    current_module = sys.modules[__name__]
    member_list: list[tuple[str, Scheme]] = inspect.getmembers(current_module, predicate=_is_scheme)
    classes: list[Scheme] = [cls for _, cls in member_list]
    return classes


@cache
def unmap_scheme(scheme: str):
    schemes = all_schemes()
    for cls in schemes:
        cls_uri_scheme = getattr(cls, "URI_SCHEME", None)
        if cls_uri_scheme is not None and cls_uri_scheme == scheme:
            return cls
    # Go through a second time, checking aliases. Keeping as two loops to always prefer primary schemes over overlap with aliases.
    for cls in schemes:
        if scheme in cls.ALIASES:
            return cls


def determine_scheme(uri: URI | str | Path) -> "Scheme | None":
    uri = URI(uri)
    scheme_cls = unmap_scheme(uri.scheme)
    if scheme_cls:
            return scheme_cls
    else:
        raise ValueError(f"Unable to identify scheme '{uri.scheme}' in uri '{uri}'")


def read_from_uri(
    uri: str,
    *args,
    **kwargs,
) -> bytes:
    scheme_cls = determine_scheme(uri)
    return scheme_cls.read(uri, *args, **kwargs)


class Scheme(ABC):
    URI_SCHEME: str
    ALIASES: list[str] = []

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .local_file_loader import LocalFileLoader
        return LocalFileLoader()

    @classmethod
    @abstractmethod
    def read(
        cls,
        uri: str,
        *args,
        **kwargs,
    ) -> bytes | str:
        ...

    @classmethod
    @abstractmethod
    def join_parts(cls, *parts: list[str]) -> str:
        ...

    @classmethod
    def get_uri_for_location(
        cls,
        location: str,
        base: str = "",
    ):
        if not location:
            raise ValueError(f"Value '{location}' is not a valid location for a URI.")
        if base:
            return f"{cls.URI_SCHEME}:{cls.join_parts(base, location)}"
        else:
            return f"{cls.URI_SCHEME}:{location}"


class LocalFileScheme(Scheme):
    URI_SCHEME = 'file'

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .local_file_loader import LocalFileLoader
        return LocalFileLoader()

    @classmethod
    def join_parts(cls, *parts: list[str]) -> str:
        return os.path.join(*parts)

    @classmethod
    def read(
        cls,
        uri: URI | str,
        base_dir: str | Path = "",
        *args,
        **kwargs,
    ) -> bytes | str:
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme '{uri.scheme}' does not match expected scheme '{cls.URI_SCHEME}'.")
        if not uri.path.startswith('/'):
            path = os.path.join(base_dir, uri.path)
        else:
            path = uri.path
        try:
            with open(path, 'r') as resource_file:
                result = resource_file.read()
        except UnicodeDecodeError:
            with open(path, 'rb') as resource_file:
                result = resource_file.read()
        return result


class ExampleScheme(LocalFileScheme):
    URI_SCHEME = 'examples'

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .local_file_loader import LocalFileLoader
        from ..resources import ResourceType
        return LocalFileLoader(resource_type=ResourceType.Example)


class DocumentationScheme(LocalFileScheme):
    URI_SCHEME = 'documentation'

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .local_file_loader import LocalFileLoader
        from ..resources import ResourceType
        return LocalFileLoader(resource_type=ResourceType.Documentation)


class PythonModuleScheme(Scheme):
    URI_SCHEME = 'py-mod'
    ALIASES = [
        "python",
        "python3",
    ]

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .code_library_loader import PythonLibraryLoader
        return PythonLibraryLoader()

    @staticmethod
    def get_module_path(mod_name) -> str | None:
        # Look for file by just looking for the file
        # First, use the loaders provided by the sys.meta_path
        head, *tail = mod_name.split('.')
        for loader in sys.meta_path:
            if not hasattr(loader, "find_spec"):
                continue
            spec = loader.find_spec(head, None)
            if not spec:
                continue

            file_paths = []
            if spec.origin is not None:
                file_paths.append(spec.origin)
            if isinstance(spec.submodule_search_locations, list):
                file_paths.extend(spec.submodule_search_locations)
            for file_path in file_paths:
                if os.path.isfile(file_path):
                    if tail:
                        file_path = os.path.split(file_path)[0]
                for submod in tail:
                    file_path = os.path.join(file_path, submod)

                if os.path.isdir(file_path):
                    file_path = os.path.join(file_path, '__init__.py')
                if not os.path.isfile(file_path) and os.path.isfile(f"{file_path}.py"):
                    file_path = f"{file_path}.py"

                if os.path.isfile(file_path):
                    return file_path

        # Try to get the file via internal Python methodology, but executes the parent modules(s)
        spec = importlib.util.find_spec(mod_name)
        if spec is not None:
            if spec.loader and hasattr(spec.loader, 'get_filename'):
                return spec.loader.get_filename()
            elif spec.origin:
                return spec.origin


    @classmethod
    def join_parts(cls, *parts: list[str]) -> str:
        return ".".join(parts)

    @classmethod
    def read(
        cls,
        uri: str,
        *args,
        **kwargs,
    ) -> bytes | str:
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme '{uri.scheme}' does not match expected scheme '{cls.URI_SCHEME}'.")

        mod_name = uri.path
        file_path = cls.get_module_path(mod_name)

        if file_path and os.path.isfile(file_path):
            with open(file_path, 'r') as resource_file:
                result = resource_file.read()
            return result

        raise FileNotFoundError(f"Unable to locate a file for module `{mod_name}`")


class S3Scheme(Scheme):
    URI_SCHEME = 's3'

    @classmethod
    def join_parts(cls, *parts: list[str]) -> str:
        return os.path.join(*parts)

    @classmethod
    def read(
        cls,
        uri: str,
        *args,
        **kwargs,
    ):
        # TODO: boto3?
        return ""


class ZipfileScheme(Scheme):
    URI_SCHEME = 'zipped-file'

    @classmethod
    def get_uri_for_location(
        cls,
        location: str,
        base: str = "",
    ):
        zipfile_path = Path(base)
        if not (
            zipfile_path.is_absolute() and zipfile_path.is_file() and str(zipfile_path).endswith(".zip")
        ):
            raise ValueError(f"Argument `base` must be an absolute path that points to a zipfile.")
        if not location:
            raise ValueError(f"Value '{location}' is not a valid location for a URI.")

        return f"{cls.URI_SCHEME}://{base}#{location}"

    @classmethod
    def join_parts(cls, *parts: list[str]) -> str:
        return os.path.join(*parts)

    @classmethod
    def read(
        cls,
        uri: str,
        *args,
        **kwargs,
    ):
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme '{uri.scheme}' does not match expected scheme '{cls.URI_SCHEME}'.")
        zipfile_path = uri.path
        inner_file = uri.fragment
        with zipfile.ZipFile(zipfile_path) as zipfile_fh:
            with zipfile_fh.open(inner_file) as inner_file_fh:
                content = inner_file_fh.read()
        try:
            return content.decode()
        except UnicodeDecodeError:
            return content


class CorpusResourceScheme(Scheme):
    URI_SCHEME = 'corpus'

    @classmethod
    def get_uri_for_location(
        cls,
        location: str,
        base: str = "",
    ):
        if not location:
            raise ValueError(f"Value '{location}' is not a valid location for a resource in a corpus.")

        return f"{cls.URI_SCHEME}:{location}"

    @classmethod
    def read(
        cls,
        uri: str,
        corpus: "Corpus",
        *args,
        **kwargs,
    ):
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme '{uri.scheme}' does not match expected scheme '{cls.URI_SCHEME}'.")
        return corpus.read_resource(uri.path)


class RCranScheme(Scheme):
    URI_SCHEME = "rcran-package"
    ALIASES = [
        "r_cran",
        "rlang",
        "r",
        "irkernel",
    ]

    local_file_cache: dict[str, str] = {}

    @classmethod
    def default_loader(cls) -> "BaseLoader":
        from .code_library_loader import RCRANSourceLoader
        return RCRANSourceLoader()

    @classmethod
    def get_uri_for_location(
        cls,
        location: str,
        base: str = "",
    ):
        if not (location and base):
            raise ValueError(f"Value '{location}' in base '{base}' is not a valid rcran-package location.")

        return f"{cls.URI_SCHEME}:{location}#{base}"

    @classmethod
    def read(
        cls,
        uri: str,
        *args,
        **kwargs,
    ):
        from .code_library_loader import RCRANLocalCache
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme '{uri.scheme}' does not match expected scheme '{cls.URI_SCHEME}'.")

        package = uri.fragment
        subpath = uri.path
        with RCRANLocalCache([package]) as cache:
            target_file = Path(cache[package]) / subpath
            with target_file.open() as source:
                content = source.read()
        return content



from .base_vector_store import VectorStore
from .chromadb_store import BaseChromaDBStore, ChromaDBLocalStore, ChromaDBServerStore, ZippedChromaDBStore


__all__ = [
    "VectorStore",
    "BaseChromaDBStore",
    "ChromaDBLocalStore",
    "ChromaDBServerStore",
    "ZippedChromaDBStore",
]



import os
from typing import List

import click

from .extract_examples import extract_examples
from .extract_documentation import extract_documentation


class BunsenCLI(click.Group):
    """
    Something something
    """

    def list_commands(self, ctx: click.Context) -> List[str]:
        # Move new and update to top of list of commands, let rest be sorted via default (alphabetical)
        commands = super().list_commands(ctx)
        commands.remove("new")
        commands.insert(0, "new")
        commands.remove("update")
        commands.insert(1, "update")
        return commands


@click.group(cls=BunsenCLI)
def cli_commands():
    """
    CLI Tooling to help build and maintain Bunsen contexts
    """
    pass


@cli_commands.command()
@click.argument("dest", required=False, type=click.Path(exists=False))
def new(dest=None):
    """
    Create a new Bunsen context via a wizard
    """
    if dest and os.path.exists(dest):
        if os.listdir(dest):
            raise click.BadParameter(
                "Destination directory exists but is not empty. Aborting.",
                param="dest",
                param_hint="destination must be empty"
            )
    click.echo("Coming soon...")


@cli_commands.command()
def update():
    """
    Check if a Bunsen context needs to be updated
    """




cli_commands.add_command(extract_examples)
cli_commands.add_command(extract_documentation)



import os
import pytest
from pathlib import Path

from beaker_bunsen.corpus.resources import Resource, ExampleResource
from beaker_bunsen.corpus.vector_stores.chromadb_store import ChromaDBLocalStore
from beaker_bunsen.corpus.loaders.local_file_loader import LocalFileLoader
from beaker_bunsen.corpus.embedders import Embedder
from beaker_bunsen.corpus.corpus import Corpus

@pytest.fixture()
def chromadb_store_path(tmp_path_factory):
    return str(tmp_path_factory.mktemp("test"))


@pytest.fixture()
def test_data_paths():
    data = Path(__file__).parent / "data"
    return [data / "examples"]


@pytest.fixture()
def chromadb_store(tmp_path_factory):
    store_path = str(tmp_path_factory.mktemp("test"))
    store = ChromaDBLocalStore(path=store_path, settings={'default_partition': "testing"})
    return store




def test_empty_example(chromadb_store):

    with pytest.raises(ValueError):
        test_resource = ExampleResource(
            uri="examples:test",
            content="",
            content_is_complete=True,
        )
        record = next(test_resource.as_records(), None)
    assert "record" not in locals()

def test_missing_description(chromadb_store):
    with pytest.raises(ValueError):
        test_resource = ExampleResource(
            uri="examples:test",
            content="""
# Code
```python
This is my code.
```
    """.strip(),
            content_is_complete=True,
        )

        record = next(test_resource.as_records(), None)
    assert "record" not in locals()

def test_bad_markdown(chromadb_store):
    from io import BytesIO
    with pytest.raises(ValueError):
        test_resource = ExampleResource(
            uri="examples:test",
            content=BytesIO(b"This cannot be parsed"),
            content_is_complete=True,
        )
        record = next(test_resource.as_records(), None)
    assert "record" not in locals()

def test_no_headings(chromadb_store):
    with pytest.raises(ValueError, ):
        test_resource = ExampleResource(
            uri="examples:test",
            content="""
*This is valid markdown*

```javascript
this.is_valid()
```

But there are no headings
""".strip(),
            content_is_complete=True,
        )

        record = next(test_resource.as_records(), None)
    assert "record" not in locals()


def test_valid_content(chromadb_store):
    content = """
# Description
This is a test example

# Code
```python
This is some python code.
```

""".strip()
    test_resource = ExampleResource(
        uri="examples:test",
        content=content,
        content_is_complete=True
    )

    record = next(test_resource.as_records(), None)

    assert record.content == content

def test_example_ingest(test_data_paths, chromadb_store):
    loader = LocalFileLoader(locations=test_data_paths)
    locations = [f"examples:{path}" for path in test_data_paths]

    corpus = Corpus(store=chromadb_store)
    corpus.ingest(locations)

    resources = list(loader.discover())
    records = chromadb_store.get_all(partition="examples")

    example_count = len(resources)
    record_count = len(records)

    assert example_count == record_count
    assert set([resource.uri for resource in resources]) == set([record.uri for record in records])
    assert all([len(record.content) > 1 for record in records])



import os
import pytest
from pathlib import Path

from beaker_bunsen.corpus.vector_stores.chromadb_store import ChromaDBLocalStore
from beaker_bunsen.corpus.corpus import Corpus
from beaker_bunsen.corpus.embedders import Embedder
from beaker_bunsen.corpus.resources import ResourceType
from beaker_bunsen.corpus.types import URI


@pytest.fixture()
def chromadb_store_path(tmp_path_factory):
    return str(tmp_path_factory.mktemp("test"))

@pytest.fixture()
def test_data_path():
    data = Path(__file__).parent / "data"
    return data / "documentation"

@pytest.fixture()
def test_data_paths(test_data_path):
    return [test_data_path]


def test_document_splitting(chromadb_store_path, test_data_path):
    # Only load the first file in the documentation directory
    fileuri = URI(f"documentation:{test_data_path / (os.listdir(test_data_path))[0]}")

    small_store = ChromaDBLocalStore(path=os.path.join(chromadb_store_path, "small"))
    small_corpus = Corpus(small_store)
    small_corpus.ingest(
        locations=[fileuri],
        embedder_map={
            ResourceType.Documentation: Embedder(
                chunk_size=100,
                chunk_overlap=15,
            ),
        }
    )

    # big_store = ChromaDBLocalStore(path=chromadb_store_path / "big")
    big_store = ChromaDBLocalStore(path=os.path.join(chromadb_store_path, "big"))
    big_corpus = Corpus(big_store)
    big_corpus.ingest(
        locations=[fileuri],
        embedder_map={
            ResourceType.Documentation: Embedder(
                chunk_size=2000,
                chunk_overlap=100,
            ),
        }
    )


    small_records = sorted(small_store.get_all(partition="documentation"), key=lambda record: record.id)
    big_records = sorted(big_store.get_all(partition="documentation"), key=lambda record: record.id)

    first_record, second_record = small_records[:2]
    record_ids = sorted(record.id for record in small_records)

    assert len(small_records) > 1
    assert len(big_records) < len(small_records)

    # Due to change in chunking, overlap is no longer assured.
    # assert first_record.content[-20:] in second_record.content[:400]  # Assert overlap feature is working

    assert first_record.uri.path == fileuri.path
    with open(fileuri.path) as rawfile:
        raw_data = rawfile.read()
        for record in small_records:
            assert record.content in raw_data
        assert raw_data.startswith(first_record.content)
    assert record_ids[0].endswith(f"{fileuri.path}:1")


def test_document_embedder(chromadb_store_path, test_data_paths):
    store = ChromaDBLocalStore(path=chromadb_store_path)
    locations = [f"file:{path}" for path in test_data_paths]
    corpus = Corpus(store)

    records_before_ingestion = store.get_all()
    corpus.ingest(locations=locations)

    ingested_records = store.get_all()
    record_ids = set(record.id for record in ingested_records)
    record_uris = set(record.uri for record in ingested_records)


    assert len(records_before_ingestion) == 0
    assert len(ingested_records) > 0
    assert len(ingested_records) == len(record_ids)
    assert len(ingested_records) > len(record_uris)
